# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import inspect, re, importlib, sys, os, logging, itertools, functools
from .common import BadArgException, NoSuchPlaceholderException, AlreadyBoundException, NotDynamicException, NoSuchVariableException
from .unroll import unroll

log = logging.getLogger(__name__)

@functools.total_ordering
class Placeholder(object):

    isplaceholder = True

    def __init__(self, name):
        self.name = name

    def __lt__(self, that):
        return self.name < that.name

    def __eq__(self, that):
        return self.name == that.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return self.name # Just import it.

    def resolvedarg(self, variant):
        return variant.paramtoarg[self]

@functools.total_ordering
class Type:

    isplaceholder = False

    def __init__(self, t):
        self.t = t

    def resolvedarg(self, variant):
        return self

    def typename(self):
        return self.t.__name__

    def discriminator(self):
        return self.typename()

    def __lt__(self, that):
        return self.t < that.t

    def __eq__(self, that):
        return self.t == that.t

    def __hash__(self):
        return hash(self.t)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.t)

    def unwrap(self):
        return self.t

class Obj:

    def __init__(self, o):
        self.o = o

    def typename(self):
        raise BadArgException(self.o)

    def discriminator(self):
        return str(self.o)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.o)

class CDef:

    def __init__(self, name, text):
        self.name = name
        self.text = text

    def __str__(self):
        return self.text

class Array:

    def __init__(self, elementtypespec, ndim):
        self.ndimtext = ", ndim=%s" % ndim if 1 != ndim else ''
        self.zeros = ', '.join(['0'] * ndim)
        self.elementtypespec = elementtypespec

    def ispotentialconst(self):
        return False

    def cparam(self, variant, name):
        elementtypename = self.elementtypespec.resolvedarg(variant).typename()
        name = 'py_' + name
        return CDef(name, "np.ndarray[np.%s_t%s] %s" % (elementtypename, self.ndimtext, name))

    def itercdefs(self, variant, name, isfuncparam):
        elementtypename = self.elementtypespec.resolvedarg(variant).typename()
        if isfuncparam:
            yield CDef(name, "cdef np.%s_t* %s = &py_%s[%s]" % (elementtypename, name, name, self.zeros))
        else:
            yield CDef(name, "cdef np.%s_t* %s" % (elementtypename, name))

    def iternestedcdefs(self, variant, undparent, dotparent, name):
        elementtypename = self.elementtypespec.resolvedarg(variant).typename()
        cname = undparent + '_' + name
        pyname = 'py_' + cname
        yield CDef(pyname, "cdef np.ndarray[np.%s_t%s] %s = %s.%s" % (elementtypename, self.ndimtext, pyname, dotparent, name))
        yield CDef(cname, "cdef np.%s_t* %s = &py_%s_%s[%s]" % (elementtypename, cname, undparent, name, self.zeros))

    def iterplaceholders(self):
        if self.elementtypespec.isplaceholder:
            yield self.elementtypespec, lambda arg: Type(arg.dtype.type)

class Scalar:

    def __init__(self, typespec):
        self.typespec = typespec

    def ispotentialconst(self):
        return self.typespec.isplaceholder

    def cparam(self, variant, name):
        typename = self.typespec.resolvedarg(variant).typename()
        return CDef(name, "np.%s_t %s" % (typename, name))

    def itercdefs(self, variant, name, isfuncparam):
        if not isfuncparam:
            typename = self.typespec.resolvedarg(variant).typename()
            yield CDef(name, "cdef np.%s_t %s" % (typename, name))

    def iternestedcdefs(self, variant, undparent, dotparent, name):
        typename = self.typespec.resolvedarg(variant).typename()
        cname = undparent + '_' + name
        yield CDef(cname, "cdef np.%s_t %s = %s.%s" % (typename, cname, dotparent, name))

    def resolvedobj(self, variant):
        return self.typespec.resolvedarg(variant).o

    def iterplaceholders(self):
        if self.typespec.isplaceholder:
            yield self.typespec, lambda arg: Type(type(arg))

class Composite:

    def __init__(self, fields):
        self.fields = sorted(fields.items())

    def cparam(self, variant, name):
        return CDef(name, name)

    def itercdefs(self, variant, name, isfuncparam):
        for field, fieldtype in self.fields:
            for cdef in fieldtype.iternestedcdefs(variant, name, name, field):
                yield cdef

    def iterplaceholders(self):
        for field, fieldtype in self.fields:
            for placeholder, resolver in fieldtype.iterplaceholders():
                yield placeholder, FieldResolver(field, resolver)

    def iternestedcdefs(self, variant, undparent, dotparent, name):
        undparent += '_' + name
        dotparent += '.' + name
        for field, fieldtype in self.fields:
            for cdef in fieldtype.iternestedcdefs(variant, undparent, dotparent, field):
                yield cdef

class FieldResolver:

    def __init__(self, field, resolver):
        self.field = field
        self.resolver = resolver

    def __call__(self, arg):
        return self.resolver(getattr(arg, self.field))

class PositionalResolver:

    def __init__(self, i, resolver):
        self.i = i
        self.resolver = resolver

    def __call__(self, args):
        return self.resolver(args[self.i])

class Variant:

    def __init__(self, decorated, paramtoarg):
        self.unbound = set(p for p in decorated.placeholders if p not in paramtoarg)
        if not self.unbound:
            self.suffix = ''.join('_' + arg.discriminator() for _, arg in sorted(paramtoarg.items()))
        self.paramtoarg = paramtoarg

    def spinoff(self, decorated, param, arg):
        if param not in decorated.placeholders:
            raise NoSuchPlaceholderException(param)
        if param in self.paramtoarg:
            raise AlreadyBoundException(param, self.paramtoarg[param].unwrap(), arg.unwrap())
        paramtoarg = self.paramtoarg.copy()
        paramtoarg[param] = arg
        return Variant(decorated, paramtoarg)

    def complete(self, decorated, args):
        if not decorated.dynamic:
            raise NotDynamicException(decorated.name)
        paramtoarg = self.paramtoarg.copy()
        for param in self.unbound:
            paramtoarg[param] = decorated.placeholdertoresolver[param](args)
        return Variant(decorated, paramtoarg)

class Decorated(object):

    pyxbld = '''from distutils.extension import Extension
import numpy as np

def make_ext(name, source):
    return Extension(name, [source], include_dirs = [np.get_include()])
'''
    template = '''# cython: language_level=3

cimport numpy as np
import cython

%(defs)s
@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def %(name)s(%(cparams)s):
%(code)s'''
    deftemplate = '''DEF %s = %r
'''
    eol = re.search(r'[\r\n]+', pyxbld).group()
    indentpattern = re.compile(r'^\s*')
    colonpattern = re.compile(r':\s*$')

    @classmethod
    def getbody(cls, pyfunc):
        lines = inspect.getsource(pyfunc).splitlines()
        getindent = lambda: cls.indentpattern.search(lines[i]).group()
        i = 0
        functionindentlen = len(getindent())
        while True:
            i += 1 # No point checking (first line of) decorator.
            if cls.colonpattern.search(lines[i]) is not None:
                break
        i += 1
        bodyindent = getindent()
        if re.search(r'=\s*LOCAL\s*$', lines[i]) is not None:
            i += 1
        return bodyindent[functionindentlen:], ''.join(line[functionindentlen:] + cls.eol for line in lines[i:])

    def __init__(self, nametotypespec, dynamic, pyfunc):
        co_varnames = pyfunc.__code__.co_varnames # The params followed by the locals.
        co_argcount = pyfunc.__code__.co_argcount
        self.paramnames = co_varnames[:co_argcount]
        self.localnames = [n for n in co_varnames[co_argcount:] if 'UNROLL' != n]
        self.constnames = []
        allnames = set(itertools.chain(self.paramnames, self.localnames))
        for name, typespec in nametotypespec.items():
            if name not in allnames:
                if not typespec.ispotentialconst():
                    raise NoSuchVariableException(name)
                self.constnames.append(name) # We'll make a DEF for it.
        self.fqmodule = pyfunc.__module__
        self.name = pyfunc.__name__
        self.bodyindent, self.body = self.getbody(pyfunc)
        # Note placeholders includes those in consts, placeholdertoresolver does not:
        self.placeholders = set()
        for typespec in nametotypespec.values():
            for param, _ in typespec.iterplaceholders():
                self.placeholders.add(param)
        self.placeholdertoresolver = {}
        for i, name in enumerate(self.paramnames):
            for placeholder, resolver in nametotypespec[name].iterplaceholders():
                if placeholder not in self.placeholdertoresolver:
                    self.placeholdertoresolver[placeholder] = PositionalResolver(i, resolver)
        self.suffixtocomplete = {}
        self.nametotypespec = nametotypespec
        self.dynamic = dynamic

    def getcomplete(self, variant):
        try:
            return self.suffixtocomplete[variant.suffix]
        except KeyError:
            self.suffixtocomplete[variant.suffix] = f = self.loadcomplete(variant)
            return f

    def loadcomplete(self, variant):
        functionname = self.name + variant.suffix
        fqmodulename = self.fqmodule + '_turbo.' + functionname
        if fqmodulename not in sys.modules:
            cparams = []
            cdefs = []
            for name in self.paramnames:
                typespec = self.nametotypespec[name]
                cparams.append(typespec.cparam(variant, name))
                cdefs.extend(typespec.itercdefs(variant, name, True))
            cdefnames = set(cdef.name for cdef in cparams)
            cdefnames.update(cdef.name for cdef in cdefs)
            for name in self.localnames:
                if name not in cdefnames:
                    typespec = self.nametotypespec[name]
                    cdefs.extend(typespec.itercdefs(variant, name, False))
            defs = []
            consts = dict([name, self.nametotypespec[name].resolvedobj(variant)] for name in self.constnames)
            for item in consts.items():
                defs.append(self.deftemplate % item)
            body = []
            unroll(self.body, body, consts, self.eol)
            body = ''.join(body)
            text = self.template % dict(
                defs = ''.join(defs),
                name = functionname,
                cparams = ', '.join(str(p) for p in cparams),
                code = ''.join("%s%s%s" % (self.bodyindent, cdef, self.eol) for cdef in cdefs) + body,
            )
            bldtext = self.pyxbld
            fileparent = os.path.join(os.path.dirname(sys.modules[self.fqmodule].__file__), self.fqmodule.split('.')[-1] + '_turbo')
            filepath = os.path.join(fileparent, functionname + '.pyx')
            bldpath = filepath + 'bld'
            existingtext = readornone(filepath)
            existingbld = readornone(bldpath)
            if text != existingtext or bldtext != existingbld:
                try:
                    os.mkdir(fileparent)
                except OSError:
                    pass
                open(os.path.join(fileparent, '__init__.py'), 'w').close()
                with open(filepath, 'w') as g:
                    g.write(text)
                    g.flush()
                with open(bldpath, 'w') as g:
                    g.write(bldtext)
                    g.flush()
                print("Compiling:", functionname, file=sys.stderr)
            importlib.import_module(fqmodulename)
        return Complete(getattr(sys.modules[fqmodulename], functionname))

    def __repr__(self):
        return "%s(<function %s>)" % (type(self).__name__, self.name)

def readornone(path):
    if os.path.exists(path):
        with open(path) as f:
            return f.read()

class Complete(object):

    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def __get__(self, instance, owner):
        return lambda *args, **kwargs: self.f(instance, *args, **kwargs)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.f)

class InstanceComplete:

    def __init__(self, instance, f):
        self.instance = instance
        self.f = f

    def __call__(self, *args, **kwargs):
        return self.f(self.instance, *args, **kwargs)

def partialorcomplete(decorated, variant):
    if variant.unbound:
        return Partial(decorated, variant)
    else:
        return decorated.getcomplete(variant)

class Partial(object):

    def __init__(self, decorated, variant):
        self.decorated = decorated
        self.variant = variant

    def todynamic(self):
        return Partial(self.decorated, self.variant, True)

    def __getitem__(self, xxx_todo_changeme):
        (param, arg) = xxx_todo_changeme
        arg = Type(arg) if isinstance(arg, type) else Obj(arg)
        return partialorcomplete(self.decorated, self.variant.spinoff(self.decorated, param, arg))

    def __call__(self, *args, **kwargs):
        return self.decorated.getcomplete(self.variant.complete(self.decorated, args))(*args, **kwargs)

    def __get__(self, instance, owner):
        return InstancePartial(instance, self.decorated, self.variant)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.decorated)

class InstancePartial:

    def __init__(self, instance, decorated, variant):
        self.instance = instance
        self.decorated = decorated
        self.variant = variant

    def __call__(self, *args, **kwargs):
        return self.decorated.getcomplete(self.variant.complete(self.decorated, (self.instance,) + args))(self.instance, *args, **kwargs)

    def __getitem__(self, xxx_todo_changeme1):
        (param, arg) = xxx_todo_changeme1
        arg = Type(arg) if isinstance(arg, type) else Obj(arg)
        variant = self.variant.spinoff(self.decorated, param, arg)
        if variant.unbound:
            return InstancePartial(self.instance, self.decorated, variant)
        else:
            return InstanceComplete(self.instance, self.decorated.getcomplete(variant))

class Decorator:

    def __init__(self, nametotypespec, dynamic):
        def wrap(spec):
            return spec if isinstance(spec, Placeholder) else Type(spec)
        def iternametotypespec(nametotypespec):
            for name, typespec in nametotypespec.items():
                if list == type(typespec):
                    elementtypespec, = typespec
                    ndim = 1
                    while list == type(elementtypespec):
                        elementtypespec, = elementtypespec
                        ndim += 1
                    typespec = Array(wrap(elementtypespec), ndim)
                elif dict == type(typespec):
                    typespec = Composite(dict(iternametotypespec(typespec)))
                else:
                    typespec = Scalar(wrap(typespec))
                yield name, typespec
        self.nametotypespec = dict(iternametotypespec(nametotypespec))
        self.dynamic = dynamic

    def __call__(self, pyfunc):
        decorated = Decorated(self.nametotypespec, self.dynamic, pyfunc)
        return partialorcomplete(decorated, Variant(decorated, {}))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import re
from io import StringIO

pattern = re.compile(r'^(\s*)for\s+UNROLL\s+in\s+range\s*\(\s*([^\s]+)\s*\)\s*:\s*$')
indentregex = re.compile(r'^\s*')
maxchunk = 0x80

def unroll(body, g, consts, eol):
    f = StringIO(body)
    buffer = []
    while True:
        line = buffer.pop(0) if buffer else f.readline()
        if not line:
            break
        m = pattern.search(line)
        if m is None:
            g.append(line)
            continue
        outerindent = m.group(1)
        variable = m.group(2)
        line = f.readline()
        m = indentregex.search(line)
        innerindent = m.group()
        body = []
        while line.startswith(innerindent):
            body.append(line)
            line = f.readline()
        buffer.append(line)
        if variable in consts:
            for _ in range(consts[variable]):
                for line in body:
                    g.append(outerindent + line[len(innerindent):])
        else:
            mask = 0x01
            while mask < maxchunk:
                g.append("%sif %s & 0x%x:%s" % (outerindent, variable, mask, eol))
                for _ in range(mask):
                    for line in body:
                        g.append(line)
                mask <<= 1
            g.append("%swhile %s >= 0x%x:%s" % (outerindent, variable, maxchunk, eol))
            for _ in range(maxchunk):
                for line in body:
                    g.append(line)
            g.append("%s%s -= 0x%x%s" % (innerindent, variable, maxchunk, eol))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

class BadArgException(Exception):

    def __init__(self, obj):
        Exception.__init__(self, obj)

class NoSuchVariableException(Exception):

    def __init__(self, name):
        Exception.__init__(self, name)

class NoSuchPlaceholderException(Exception):

    def __init__(self, param):
        Exception.__init__(self, param)

class AlreadyBoundException(Exception):

    def __init__(self, param, current, given):
        Exception.__init__(self, param, current, given)

class NotDynamicException(Exception):

    def __init__(self, name):
        Exception.__init__(self, name)



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division
import unittest, numpy as np, logging, time
from .leaf import turbo, T

log = logging.getLogger(__name__)

def pysum(n, x, y, out):
    for i in range(n):
        out[i] = x[i] + y[i]

def npsum(n, x, y, out):
    out[:n] = x[:n]
    out[:n] += y[:n]

@turbo(i = np.uint32, n = np.uint32, x = [np.float32], y = [np.float32], out = [np.float32])
def tsum(n, x, y, out):
    for i in range(n):
        out[i] = x[i] + y[i]

@turbo(types = dict(i = np.uint32, n = np.uint32, x = [T], y = [T], out = [T]), dynamic = True)
def gsum(n, x, y, out):
    for i in range(n):
        out[i] = x[i] + y[i]

class Cls:

    @turbo(types = dict(self = {}, i = np.uint32, n = np.uint32, x = [T], y = [T], out = [T]), dynamic = True)
    def oogsum(self, n, x, y, out):
        for i in range(n):
            out[i] = x[i] + y[i]

class TestTurbo(unittest.TestCase):

    def test_works(self):
        n = 100000
        x = np.arange(n, dtype = np.float32)
        y = np.arange(n, dtype = np.float32) * 2
        expected = np.empty(n, dtype = np.float32)
        pysum(n, x, y, expected)
        for task in npsum, tsum, gsum[T, np.float32], gsum, Cls().oogsum, Cls().oogsum[T, np.float32]:
            actual = np.empty(n, dtype = np.float32)
            task(n, x, y, actual)
            self.assertTrue(np.array_equal(expected, actual))

class TestSpeed(unittest.TestCase):

    ntomaxreltime = {100: 1.2, 1000: 1.2} # I guess in small array case we see some setup overhead.

    def test_fastenough(self):
        trials = 20
        outliers = 2
        meanof = lambda v: sum(v) / len(v)
        nandtasktotimes = {}
        for n in range(2, 6):
            n = 10 ** n
            log.info("n: %s", n)
            x = np.arange(n, dtype = np.float32)
            y = np.arange(n, dtype = np.float32) * 2
            for task in npsum, tsum, gsum[T, np.float32]:
                log.info("task: %s", task)
                out = np.empty(n, dtype = np.float32)
                times = []
                for _ in range(trials):
                    t = time.time()
                    task(n, x, y, out)
                    times.append((time.time() - t) * 1000)
                times.sort()
                log.info("trials: %s", ' '.join("%.6f" % q for q in times))
                nandtasktotimes[n, task] = times[outliers:-outliers]
        fails = []
        for (n, task), times in nandtasktotimes.items():
            if npsum != task:
                maxreltime = self.ntomaxreltime.get(n, 1)
                reltime = meanof(times) / meanof(nandtasktotimes[n, npsum])
                if reltime > maxreltime:
                    fails.append("(n = %r, task = %r, reltime = %r, maxreltime = %r)" % (n, task, reltime, maxreltime))
        self.assertEqual([], fails)

@turbo(n = np.uint32, acc = np.uint32)
def triple(n):
    acc = 0
    for UNROLL in range(n):
        acc += 3
    return acc

class TestUnroll(unittest.TestCase):

    def test_unroll(self):
        self.assertEqual(21, triple(7))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np, unittest
from .leaf import turbo, T, U, X

@turbo(types = dict(x = [T], y = [U], n = np.uint32), dynamic = True)
def addxtoy(x, y, n):
    while n:
        n -= 1
        y[n] += x[n]

@turbo(types = dict(x = X, y = X), dynamic = True)
def add(x, y):
    return x + y

@turbo(types = dict(x = [T], y = [T]), dynamic = True)
def noinfer(x, y):
    pass

class TestInfer(unittest.TestCase):

    def test_works(self):
        x = np.arange(10, dtype = np.int32) + 1
        y = np.arange(10, dtype = np.float32)
        addxtoy(x, y, 5)
        self.assertEqual(list(range(1, 11)), list(x))
        self.assertEqual([1, 3, 5, 7, 9, 5, 6, 7, 8, 9], list(y))

    def test_works2(self):
        self.assertEqual(11, add(5, 6))

    def test_failure(self):
        x = np.arange(10, dtype = np.int32)
        y = np.arange(10, dtype = np.float32) # Type not used.
        try:
            noinfer(x, y)
            self.fail("Expected value error.")
        except ValueError as e:
            self.assertEqual(("Buffer dtype mismatch, expected 'int32_t' but got 'float'",), e.args)



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np, unittest
from .leaf import turbo, T

@turbo(types = dict(a = [[T]]), dynamic = True)
def multidim(a):
    return a[0], a[1], a[2], a[3]

class TestMultiDim(unittest.TestCase):

    def test_works(self):
        a = np.empty((2, 2), dtype = np.float32)
        a[0, 0] = 1
        a[0, 1] = 2
        a[1, 0] = 3
        a[1, 1] = 4
        self.assertEqual((1, 2, 3, 4), multidim(a))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import unittest, numpy as np
from .leaf import turbo, T

@turbo(v = [T])
def f(v):
    pass

@turbo(x = T)
def f2(x):
    pass

class TestExact(unittest.TestCase):

    def test_exact(self):
        dtypes = (np.int8, np.int16, np.int32, np.int64,
            np.uint8, np.uint16, np.uint32, np.uint64,
            np.float32, np.float64)
        for t in dtypes:
            for u in dtypes:
                g = f[T, t]
                v = np.empty(10, dtype = u)
                if t == u:
                    g(v)
                else:
                    try:
                        g(v)
                        self.fail('Expected value error.')
                    except ValueError:
                        pass
                g2 = f2[T, t]
                x = u(10)
                g2(x) # All conversions allowed, including narrowing.



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

from pyrboimpl.leaf import turbo, T, U, V, W, X, Y, Z, generic, LOCAL
from pyrboimpl.common import BadArgException, NoSuchVariableException, NoSuchPlaceholderException, AlreadyBoundException

assert BadArgException
assert NoSuchVariableException
assert NoSuchPlaceholderException
assert AlreadyBoundException

assert turbo
assert T
assert U
assert V
assert W
assert X
assert Y
assert Z
assert generic
assert not LOCAL



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import initnative, pyrboimpl
from .pyrboimpl import Placeholder, Partial, Type, Obj, Decorator
del initnative

globals().update([p.name, p] for p in (Placeholder(chr(i)) for i in range(ord('T'), ord('Z') + 1)))

def turbo(**kwargs):
    if 'types' not in kwargs:
        kwargs = dict(types = kwargs)
    nametotypespec = kwargs['types']
    dynamic = kwargs.get('dynamic', False)
    return Decorator(nametotypespec, dynamic)

class ClassVariant:

    @classmethod
    def create(cvcls, cls):
        placeholders = set()
        for member in cls.__dict__.values():
            if isinstance(member, Partial):
                placeholders.update(member.decorated.placeholders)
        return cvcls(cls.__name__, placeholders, {})

    def __init__(self, basename, placeholders, paramtoarg):
        self.basename = basename
        self.placeholders = placeholders
        self.paramtoarg = paramtoarg

    def spinoff(self, param, arg):
        if param not in self.placeholders:
            raise pyrboimpl.common.NoSuchPlaceholderException(param)
        if param in self.paramtoarg:
            raise pyrboimpl.common.AlreadyBoundException(param, self.paramtoarg[param].unwrap(), arg.unwrap())
        paramtoarg = self.paramtoarg.copy()
        paramtoarg[param] = arg
        return ClassVariant(self.basename, self.placeholders, paramtoarg)

class basegeneric(type):

    def __getitem__(cls, xxx_todo_changeme):
        (param, arg) = xxx_todo_changeme
        arg = Type(arg) if isinstance(arg, type) else Obj(arg)
        variant = cls.turbo_variant.spinoff(param, arg)
        members = {}
        for name, member in cls.__dict__.items():
            if isinstance(member, Partial) and param in member.variant.unbound:
                member = member[param, arg.unwrap()]
            members[name] = member
        members['turbo_variant'] = variant
        words = [variant.basename]
        for param in sorted(variant.placeholders):
            words.append(variant.paramtoarg[param].discriminator() if param in variant.paramtoarg else '?')
        return basegeneric('_'.join(words), cls.__bases__, members)

class generic(basegeneric):

    def __new__(self, name, bases, members):
        cls = basegeneric.__new__(self, name, bases, members)
        cls.turbo_variant = ClassVariant.create(cls)
        return cls

LOCAL = None



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import unittest, numpy as np
from .leaf import turbo

@turbo(
    i = np.uint32,
    n = np.uint32,
    x = [np.float32],
    y = [np.float32],
    out = [np.float32],
)
def tsum(
    n,
    x,
    y,
    out,
):
    for i in range(n):
        out[i] = x[i] + y[i]

class TestTurbo(unittest.TestCase):

    def test_works(self):
        n = 100000
        x = np.arange(n, dtype = np.float32)
        y = np.arange(n, dtype = np.float32) * 2
        actual = np.empty(n, dtype = np.float32)
        for task in tsum,:
            task(n, x, y, actual)
            self.assertTrue(np.array_equal(x + y, actual))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import unittest, numpy as np
from .leaf import turbo, X, T, U, Z, generic, LOCAL
from .common import NoSuchPlaceholderException, AlreadyBoundException

class My:

    def __init__(self, x):
        self.x = x

    @turbo(self = dict(x = np.int8), y = np.int8, z = np.int8)
    def plus(self, y):
        self_x = LOCAL
        z = self_x + y
        return z

    @turbo(types = dict(self = dict(x = X), y = X, z = X), dynamic = True)
    def plus2(self, y):
        self_x = LOCAL
        z = self_x + y
        return z

class My2:

    def __init__(self, x):
        self.x = x

    @turbo(self = dict(x = np.int8), y = np.int8, z = np.int8)
    def plus(self, y):
        self_x = LOCAL
        z = self_x - y
        return z

@turbo(obj = dict(field = int))
def fieldlocal(obj):
    obj_field = 6
    if False:
        obj.field = obj_field

class TestOO(unittest.TestCase):

    def test_works(self):
        my = My(5)
        self.assertEqual(11, my.plus(6))
        self.assertEqual(11, my.plus2(6))

    def test_works2(self):
        my = My2(5)
        try:
            self.assertEqual(-1, my.plus(6))
            raise Exception('You fixed a bug!')
        except AssertionError:
            pass

    def test_fieldlocal(self):
        class Obj: pass
        obj = Obj()
        obj.field = 5
        fieldlocal(obj)
        self.assertEqual(5, obj.field)

class Buf(metaclass=generic):

    def __init__(self, u):
        self.u = u

    @turbo(types = dict(self = dict(u = [T]), i = np.uint32, j = np.uint32, v = U), dynamic = True)
    def fillpart(self, i, j, v):
        self_u = LOCAL
        while i < j:
            self_u[i] = v
            i += 1

class TestBuf(unittest.TestCase):

    def test_works(self):
        t = np.uint16
        u = np.int32
        tbuf = Buf[T, t]
        ubuf = Buf[U, u]
        names = ['Buf', 'Buf_uint16_?', 'Buf_?_int32', 'Buf_uint16_int32', 'Buf_uint16_int32']
        for bufcls in Buf, tbuf, ubuf, tbuf[U, u], ubuf[T, t]:
            self.assertEqual(names.pop(0), bufcls.__name__)
            v = np.zeros(10, dtype = t)
            buf = bufcls(v)
            buf.fillpart(4, 6, 5)
            self.assertEqual([0, 0, 0, 0, 5, 5, 0, 0, 0, 0], list(v))
        try:
            Buf[Z, None]
            self.fail('Expected no such placeholder.')
        except NoSuchPlaceholderException as e:
            self.assertEqual((Z,), e.args)
        try:
            tbuf[T, TestBuf]
            self.fail('Expected already bound.')
        except AlreadyBoundException as e:
            self.assertEqual((T, t, TestBuf), e.args)



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division # Required by tests script.
import numpy as np, unittest
from .leaf import turbo, T

def pytrue(x, y):
    return x / y

def pyfloor(x, y):
    return x // y

@turbo(x = T, y = T)
def turbotrue(x, y):
    return x / y

@turbo(x = T, y = T)
def turbofloor(x, y):
    return x // y

class TestDivision(unittest.TestCase):

    def test_floats(self):
        for truediv in pytrue, turbotrue[T, np.float32]:
            self.assertEqual(1.5, truediv(4.5, 3))
            self.assertEqual(-1.5, truediv(-4.5, 3))
        for floordiv in pyfloor, turbofloor[T, np.float32]:
            self.assertEqual(1, floordiv(4.5, 3))
            self.assertEqual(-2, floordiv(-4.5, 3))

    def test_ints(self):
        # Native function same as for floats:
        self.assertEqual(1.5, pytrue(3, 2))
        self.assertEqual(-1.5, pytrue(-3, 2))
        # turbo function does round-to-zero:
        self.assertEqual(1, turbotrue[T, np.int32](3, 2))
        self.assertEqual(-1, turbotrue[T, np.int32](-3, 2))
        # Native function same as for floats:
        self.assertEqual(1, pyfloor(3, 2))
        self.assertEqual(-2, pyfloor(-3, 2))
        # turbo function does round-to-zero:
        self.assertEqual(1, turbofloor[T, np.int32](3, 2))
        self.assertEqual(-1, turbofloor[T, np.int32](-3, 2))



# Copyright 2015, 2016, 2017 Andrzej Cichocki

# This file is part of pyrbo.
#
# pyrbo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyrbo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyrbo.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np, unittest
from .leaf import turbo, T, X, Y
from .common import BadArgException

y = None

@turbo(x = T, y = Y)
def turbotuple(x):
    return x, y

@turbo(v = [X])
def arrayof(v):
    pass

class TestValueArg(unittest.TestCase):

    def test_works(self):
        self.assertEqual((-5, 6), turbotuple[T, np.int32][Y, 6](-5))

    def test_arrayof(self):
        try:
            arrayof[X, 100]
            self.fail("Expected bad arg.")
        except BadArgException as e:
            self.assertEqual((100,), e.args)



import setuptools

def long_description():
    with open('README.md') as f:
        return f.read()

setuptools.setup(
        name = 'pyrbo',
        version = '3',
        description = 'Python JIT compiler for near-native performance of low-level arithmetic',
        long_description = long_description(),
        long_description_content_type = 'text/markdown',
        url = 'https://github.com/combatopera/pyrbo',
        author = 'Andrzej Cichocki',
        packages = setuptools.find_packages(),
        py_modules = ['pyrbo'],
        install_requires = ['numpy', 'cython', 'nativecommon'],
        package_data = {'': ['*.pxd', '*.pyx', '*.pyxbld', '*.arid', '*.aridt']},
        scripts = [])
